// The FlipReverseRotate Library provides three functions to flip and/or
// reverse a bitmap.  You can choose which approach you'd like to use from the
// three methods to flip or reverse a bitmap:  ScanLine, CopyRect, StretchBlt.
//
// A "Flip" operation takes the top of an image to the bottom and the bottom of
// the image to the top.  It is a reflection along a horizontal line in the
// middle of an image.
//
// A "Reverse" operation takes the left of an image to the right and the right
// of the image to the left.  It is a reflection along a vertical line in the
// middle of an image.
//
// Any Flip/Reverse operation is commutative, i.e., the flip and reverse can
// be performed in any order to get the same result.  A flip followed by a
// reverse is the same as a reverse followed by a flip.
//
// A "rotate" operation spins an image 0, 90, 180 or 270 degrees around an
// axis in the center of the image.
//
// A flip/reverse operation along with a rotation is not commutative in general.
// A flip followed by a rotation will not always result in the same image as a
// rotation followed by a flip.  The rotation here ALWAYS follows any flip and/or
// reversal.
//
// The examples here are intended for use with bitmaps that have 24 bits/pixel.
// Palettes may be lost on 256-color bitmaps.
//
// Copyright (C) 1998, Earl F. Glynn.  All Rights Reserved.
// May be used freely for non-comercial use.

UNIT FlipReverseRotateLibrary;

INTERFACE

  USES
  Dialogs,
    Graphics;  // TBitmap

  // Flip/Reverse functions by Method


  FUNCTION FlipReverseCopyRect(CONST Flip, Reverse:  BOOLEAN;
                               CONST Bitmap:  TBitmap):  TBitmap;


IMPLEMENTATION

  USES
    Classes,    // Rect
    SysUtils;   // Exception

  CONST
    MaxPixelCount = 65536;   // or some other arbitrarily large value

  TYPE
    EBitmapError = CLASS(Exception);





  //////////////////////////////////////////////////////////////////////////////

  // This function implements a suggestion by David Ullrich in a July 25, 1997
  // post to comp.lang.pascal.delphi.misc.
  //
  // The Graphics.PAS unit shows that CopyRect calls the Windows StretchBlt API
  // function.
  FUNCTION FlipReverseCopyRect(CONST Flip, Reverse:  BOOLEAN;
                               CONST Bitmap:  TBitmap):  TBitmap;
    VAR
      Bottom:  INTEGER;
      Left  :  INTEGER;
      Right :  INTEGER;
      Top   :  INTEGER;
  BEGIN
    RESULT := TBitmap.Create;
    RESULT.Width       := Bitmap.Width;
    RESULT.Height      := Bitmap.Height;
    RESULT.PixelFormat := Bitmap.PixelFormat;

    // Flip Top to Bottom
    IF   Flip
    THEN BEGIN
      // Unclear why extra "-1" is needed here.
      Top    := Bitmap.Height-1;
      Bottom := -1
    END
    ELSE BEGIN
      Top    := 0;
      Bottom := Bitmap.Height
    END;

    // Reverse Left to Right
    IF   Reverse
    THEN BEGIN
      // Unclear why extra "-1" is needed here.
      Left  := Bitmap.Width-1;
      Right := -1;
    END
    ELSE BEGIN
      Left  := 0;
      Right := Bitmap.Width;
    END;

    RESULT.Canvas.CopyRect(Rect(Left,Top, Right,Bottom),
                           Bitmap.Canvas,
                           Rect(0,0, Bitmap.Width,Bitmap.Height));
  END {FlipReverseCopyRect};


END.
